#include "common/cpu_anim.h"
#include "common/cpu_bitmap.h"
#include <iostream>
#include <vector>
#include <algorithm>
#include <stdexcept>
using namespace std;
#include <getopt.h>
#include "cpu_timer.h"
#include "complex.h"


// ==================================================================
// global variables
// ==================================================================
int IMAGE_WIDTH = 1024;
int IMAGE_HEIGHT = 1024;
bool view_flag = false;
int iterations = 200;
REAL constant_real = 0.3;
REAL constant_imag = 0.5;

const REAL xmin = -2.4;
const REAL xmax = 2.4;
const REAL ymin = -1.5;
const REAL ymax = 1.5;
const REAL NORM_THRESOLD = 4.0;

// ==================================================================
// program command line arguments
// ==================================================================
static struct option long_options[] = {
		{"width"		, required_argument, 0, 'w'},
		{"height"		, required_argument, 0, 'h'},
		{"view"			, 		no_argument, 0, 'v'},
		{"iterations"	, required_argument, 0, 'i'},
		{"constant-real", required_argument, 0, 'r'},
		{"constant-imag", required_argument, 0, 's'},
		{0,0,0,0}
};

/**
 * determine if given point is in Julia Set
 *
 */
REAL julia(int x, int y, int max_iter) {
	REAL r = x * (REAL)(xmax - xmin)/IMAGE_WIDTH + xmin;
	REAL i = y * (REAL)(ymax - ymin)/IMAGE_HEIGHT + ymin;

	aComplex z(r, i);
	aComplex c(constant_real, constant_imag);

	for (int k=0; k<max_iter; k++) {
		z = z * z + c;
		if (z.norm() > NORM_THRESOLD)  {
			return NORM_THRESOLD;
		}
	}

	return z.norm();
}


void kernel(unsigned char *ptr, int max_iter){

	for (int y=0; y<IMAGE_HEIGHT; y++) {
		for (int x=0; x<IMAGE_WIDTH; x++) {

			int offset = x + y * IMAGE_WIDTH;
			REAL juliaValue = julia(x, y, max_iter );

			int  color = round(juliaValue * 255);
			color = 255 - color;
			ptr[offset*4 + 0] = color;
			ptr[offset*4 + 1] = color;
			ptr[offset*4 + 2] = color;
			ptr[offset*4 + 3] = 255;
		}
	}
}


int main(int argc, char *argv[]) {
	int option_index;
	while (true) {
		option_index = 0;
		int c = getopt_long(argc, argv, "w:h:i:r:s:v", long_options, &option_index);
		if (c == -1) break;

		switch(c) {

		case 'w':
			IMAGE_WIDTH = atoi(optarg);
			if ((IMAGE_WIDTH <= 128) || (IMAGE_WIDTH >= 8192)) {
				throw std::invalid_argument("bad image width : 128..8192");
			}
			break;

		case 'h':
			IMAGE_HEIGHT = atoi(optarg);
			if ((IMAGE_HEIGHT <= 128) || (IMAGE_HEIGHT >= 8192)) {
				throw std::invalid_argument("bad image height : 128..8192");
			}
			break;

		case 'v':
			view_flag = true;
			break;

		case 'i':
			iterations = atoi(optarg);
			if ((iterations <= 100) || (iterations >= 10000000)) {
				throw std::invalid_argument("bad number of iterations : 100..10000000");
			}
			break;

		case 'r':
			constant_real = atof(optarg);
			break;

		case 's':
			constant_imag = atof(optarg);
			break;
		}
	}

	cout << "image_width=" << IMAGE_WIDTH << endl;
	cout << "image_height=" << IMAGE_HEIGHT << endl;
	cout << "iterations=" << iterations << endl;
	cout << "constant_real=" << constant_real << endl;
	cout << "constant_imag=" << constant_imag << endl;

	CPUBitmap bitmap(IMAGE_WIDTH, IMAGE_HEIGHT);
	unsigned char *ptr = bitmap.get_ptr();


	CPUTimer c_timer;

	c_timer.start();
	kernel(ptr, iterations);
	c_timer.stop();
	cout << "cpu time(ms)=" << c_timer << endl;

	if (view_flag) {
		bitmap.display_and_exit();
	}

	return 0;
}

